import os
import ssl
import pwd
import time
import socket
import ipaddr
import pickle
import urllib2
import httplib
import subprocess

from dnsutil import DNSCache
import se_dns

DNS_CACHE = DNSCache()

_LOCAL_IP_ADDRESSES_CACHE = {}


def local_ip_addresses(cache_fn, include_external=False, use_cached=False,
                       logger=None):
    """Get all the IP addresses in use on this server.

    This calls an external process, so should be used lightly.
    """
    try:
        return _LOCAL_IP_ADDRESSES_CACHE[include_external]
    except KeyError:
        pass
    if use_cached:
        try:
            cache_good = time.time() - os.stat(cache_fn).st_mtime < 60 * 60 * 24
        except (OSError, EOFError, IOError) as e:
            if logger:
                logger.debug("Not using cache: %s", e)
            cache_good = False
        if cache_good:
            with open(cache_fn) as cache_file:
                try:
                    result = pickle.load(cache_file)
                except (OSError, EOFError, IOError) as e:
                    if logger:
                        logger.warn("Bad local IP address cache; replacing.")
                else:
                    if logger:
                        logger.debug("Using cache for local IP addresses.")
                    _LOCAL_IP_ADDRESSES_CACHE[include_external] = result
                    return result
        elif logger:
            logger.debug("Not using cache for local IP addresses.")
    ips = set()
    # XXX Once we are using psutil 3.0+, we should switch to using that
    # XXX rather than this fairly fragile subprocess.
    # XXX e.g. for IPv4: [addr.address
    # XXX                 for key, addrs in psutil.net_if_addrs().iteritems()
    # XXX                 if key != 'lo'
    # XXX                 for addr in addrs if addr.family==socket.AF_INET]
    # XXX and IPv6 the same, but socket.AF_INET6.
    cmd_templ = ("/sbin/ifconfig | grep '[[:space:]]inet[[:space:]]' "
                 "| grep -v '127.0.0.1' | "
                 "awk '{ print $2 }'")
    ip_list = subprocess.Popen(
        cmd_templ, shell=True, stdout=subprocess.PIPE,
        stdin=subprocess.PIPE).communicate()[0].strip()
    if ip_list:
        ips.update([ip for ip in ip_list.split()])
    cmd_templ = ("/sbin/ifconfig | grep '[[:space:]]inet6[[:space:]]' | "
                 "grep -i global | awk '{ print $2 }'")
    ip_list = subprocess.Popen(
        cmd_templ, shell=True, stdout=subprocess.PIPE,
        stdin=subprocess.PIPE).communicate()[0].strip()
    if ip_list:
        ips.update([ip.split("/", 1)[0] for ip in ip_list.split()])
    if include_external:
        req = urllib2.Request(include_external,
                              headers={"User-Agent": "se_dns/%s " %
                                       se_dns.__version__})
		# XXX We should use requests.
        try:
            local_ip = urllib2.urlopen(req, timeout=5).read().strip()
        except (socket.error, socket.timeout, urllib2.URLError,
                httplib.HTTPException) as e:
            if logger:
                if hasattr(e, "errno") and e.errno == 101:
                    # Network issue. Nothing we can do about this.
                    logger.info("Unable to retrieve external IP: %s", e)
                else:
                    logger.warn("Unable to retrieve external IP: %s", e)
        else:
            if local_ip.startswith("::ffff:"):
                # This is junk that nginx adds to IPv4 addresses.
                local_ip = local_ip.rsplit(":", 1)[1]
            ips.add(local_ip)
    # No longer add data to ips, we can convert it to a list to serialize it
    ips = list(ips)
    if logger:
        logger.info("Local IP addresses: %s", ", ".join(ips))
    adjust_cache = not os.path.exists(cache_fn)
    try:
        with open(cache_fn, "wb") as cache_file:
            pickle.dump(ips, cache_file, pickle.HIGHEST_PROTOCOL)
    except (IOError, OSError) as e:
        if logger:
            logger.warn("Unable to cache IP list: %s", e)
    else:
        # XXX This should not have the users hard-coded.
        if adjust_cache:
            # The cache might be created by the API (www-data), local_scan
            # (Debian-exim), or setup/update (root).  Ensure that everyone
            # can work with it.  The www-data user is in the Debian-exim
            # group.
            try:
                os.chmod(cache_fn, 0o660)
            except OSError as e:
                if logger:
                    logger.info("Unable to set permissions of cache: %s", e)
            try:
                os.chown(cache_fn, os.geteuid(),
                         pwd.getpwnam("Debian-exim").pw_gid)
            except (OSError, IOError) as e:
                if logger:
                    logger.info("Unable to set owner/group of cache: %s", e)
    result = list(set(ips))
    _LOCAL_IP_ADDRESSES_CACHE[include_external] = result
    return result


def name_to_ip(name, prefer_ipv6=True):
    """Basically like socket.gethostbyname(), except that it also supports
    IPv6.  The caller is responsible for setting an appropriate timeout and
    handling any errors."""
    ipv4 = None
    ipv6 = None
    # XXX We really want a timeout on this somehow.  Can we do this
    # XXX differently to make that possible?  We don't want to use
    # XXX signals, because that causes issues with Exim.  We could use
    # XXX a subprocess or thread, although that's a bit ugly.
    for (family, socktype, proto, canonname,
         sockaddr) in socket.getaddrinfo(name, None):
        if family == socket.AF_INET6:
            ipv6 = sockaddr[0]
            if prefer_ipv6:
                return ipv6
        elif family == socket.AF_INET:
            ipv4 = sockaddr[0]
    return ipv4 or ipv6


def hosts_equal(host1, host2, cache_fn, skip_getaddrinfo=False, logger=None,
                external_link=None):
    """Return True if these are the same physical machine."""
    if host1 == host2:
        return True
    # Currently we just check if the hosts resolve to the same IP, with
    # some special-casing for localhost.
    hostname = socket.gethostname()
    host1_is_ip = False
    host2_is_ip = False
    try:
        ip1 = ipaddr.IPAddress(host1)
    except ValueError:
        if host1 == "localhost":
            host1 = hostname
        ip1 = None
    else:
        if ip1.is_loopback or (hasattr(ip1, "site_local") and ip1.site_local):
            host1 = hostname
        else:
            host1_is_ip = True
    try:
        ip2 = ipaddr.IPAddress(host2)
    except ValueError:
        if host2 == "localhost":
            host2 = hostname
        ip2 = None
    else:
        if ip2.is_loopback or (hasattr(ip2, "site_local") and ip2.site_local):
            host2 = hostname
        else:
            host2_is_ip = True
    # Try DNS lookups first.
    host1_ips = []
    host1_has_ipv6 = False
    if not host1_is_ip:
        host1_ips.extend(DNS_CACHE.lookup(host1, "A", exact=True))
        host1_aaaa = DNS_CACHE.lookup(host1, "AAAA", exact=True)
        if host1_aaaa:
            host1_ips.extend(host1_aaaa)
            host1_has_ipv6 = True
    if ip1:
        host1_ips.append(str(ip1))
    if host1 == hostname:
        host1_ips.extend(local_ip_addresses(cache_fn, include_external=external_link,
                                            use_cached=True, logger=logger))
    host2_ips = []
    if not host2_is_ip:
        host2_ips.extend(DNS_CACHE.lookup(host2, "A", exact=True))
        if host1_has_ipv6:
            # If host1 doesn't have any AAAA records, then there's no
            # point trying to find any matching ones from host2.
            host2_ips.extend(DNS_CACHE.lookup(host2, "AAAA", exact=True))
    if ip2:
        host2_ips.append(str(ip2))
    if host2 == hostname:
        host2_ips.extend(local_ip_addresses(cache_fn, include_external=external_link,
                                            use_cached=True, logger=logger))
    if set(host1_ips).intersection(set(host2_ips)):
        return True
    if skip_getaddrinfo:
        return False
    # Fall back to getaddrinfo.
    try:
        if host1_is_ip:
            host1_ip = host1
        else:
            host1_ip = name_to_ip(host1)
        if host2_is_ip:
            host2_ip = host2
        else:
            host2_ip = name_to_ip(host2)
        return host1_ip == host2_ip
    except (socket.error, UnicodeError):
        return False


def host_in_list(host, host_list, skip_getaddrinfo=False, logger=None,
                 skip_external=False):
    """Return True if the host is equal to any of the ones in the specified
    list.
    """
    return any(hosts_equal(host, other_host, skip_getaddrinfo, logger,
                           skip_external) for other_host in host_list)


def urlopen(url, data=None, timeout=None):
    if hasattr(ssl, '_create_unverified_context'):
        ssl_context = ssl._create_unverified_context()
        return urllib2.urlopen(url, data, timeout, context=ssl_context)
    else:
        return urllib2.urlopen(url)



"""A wrapper around dnspython.

The wrapper adds two sets of functionality.

Firstly, it provides a single cache object, whenever possible, that is
used throughout the lifetime of the process.  This maximises the use of
the cached information, without having to pass a cache object around to
different parts of the code.  The code handles the case where the
timeout for the cache is different, without the caller needing to
manage that.

Secondly, it includes a transparent shortcut for various DNSBL lookups
where we have a mirror of the list, and the list provides only a
"yes listed" or "no not listed" response (or if it provides more
responses, we are not interested in them).  It does this by looking up
the value in a "combined" result, which provides results for many
different lists, for the first lookup, and then uses that information
for all subsequent checks.  From the caller's point of view, the
original list is being checked.

This code was originally based on the dnscache in SpamBayes.
"""

from __future__ import absolute_import

import os
import json
import struct
import logging

import dns.resolver
import dns.exception
import dns.rdatatype
import dns.rdataclass
import dns.reversename


DNSException = dns.exception.DNSException
EmptyLabel = dns.name.EmptyLabel
LabelTooLong = dns.name.LabelTooLong

# Default config_path hardcoded
CONFIG_PATH = "/etc/combined_lists.json"


class Cache(object):
    """Provide a simple-to-use interface to DNS lookups, which caches the
    results in memory."""
    timeout_log_level = "info"

    def __init__(self, dnsserver=None, dnstimeout=10,
                 minttl=0, cachefile=""):
        self.logger = logging.getLogger("se-dns")
        # We don't use the cachefile argument, but it may be provided.
        # XXX We can probably just drop cachefile now.
        if cachefile:
            self.logger.warn("Caching to file is not supported.")

        # Some servers always return a TTL of zero. In those cases, turning
        # this up a bit is probably reasonable.
        self.minTTL = minttl

        self.queryObj = dns.resolver.Resolver()
        if dnsserver:
            self.queryObj.nameservers = [dnsserver]

        # How long to wait for the server (in seconds).
        # dnspython has a "timeout" value (for each nameserver) and a
        # "lifetime" value (for the complete query).  We're happy with the
        # 2 second default timeout, but want to limit the overall query.
        self.queryObj.lifetime = dnstimeout

        # Use the package's caching system.
        self.queryObj.cache = dns.resolver.Cache()
        # Except that we also want to cache failures, because we are
        # generally short-lived, and sometimes errors are slow to generate.
        self.failures = {}

    def lookup(self, question, qtype="A", ctype="IN", exact=False):
        """Do an actual lookup.  'question' should be the hostname or IP to
        query, and 'qType' should be the type of record to get (e.g. TXT,
        A, AAAA, PTR)."""
        rdtype = dns.rdatatype.from_text(qtype)
        rdclass = dns.rdataclass.from_text(ctype)
        try:
            return self.failures[question, rdtype, rdclass]
        except KeyError:
            pass
        reply = self.queryObj.cache.get((question, rdtype, rdclass))
        if not reply:
            try:
                reply = self.queryObj.query(question, rdtype, rdclass)
            except dns.resolver.NXDOMAIN:
                # This is actually a valid response, not an error condition.
                self.failures[question, rdtype, rdclass] = []
                return []
            except dns.exception.Timeout:
                # This may change next time this is run, so warn about that.
                log_method = getattr(self.logger, self.timeout_log_level)
                log_method("%s %s lookup timed out.", question, qtype)
                self.failures[question, rdtype, rdclass] = []
                return []
            except (dns.resolver.NoAnswer, dns.resolver.NoNameservers) as e:
                if qtype not in ("MX", "AAAA", "TXT"):
                    # These indicate a problem with the nameserver.
                    self.logger.debug("%s %s lookup failed: %s", question,
                                      qtype, e)
                self.failures[question, rdtype, rdclass] = []
                return []
            except (ValueError, IndexError) as e:
                # A bad DNS entry.
                self.logger.warn("%s %s lookup failed: %s", question, qtype,
                                 e)
                self.failures[question, rdtype, rdclass] = []
                return []
            except struct.error as e:
                # A bad DNS entry.
                self.logger.warn("%s %s lookup failed: %s", question, qtype,
                                 e)
                self.failures[question, rdtype, rdclass] = []
                return []
        self.queryObj.cache.put((question, rdtype, rdclass), reply)
        if exact:
            return [i.to_text() for sublist in
                    (answer.to_rdataset().items
                     for answer in reply.response.answer
                     if answer.rdtype == rdtype and
                     answer.rdclass == rdclass) for i in sublist]
        return [i.to_text()
                for i in reply.response.answer[0].to_rdataset().items]


class _DNSCache(Cache):
    """Like the parent, but also knows about the combined DNSBL and URLBL
    that enable checking many DNSBL / URLBL with a single lookup.

    Use is transparent to the caller - i.e. they use the normal name for
    the list, and this class does the work of deciding whether to instead
    query the combined list.
    """

    def __init__(self, config_path=CONFIG_PATH):
        super(_DNSCache, self).__init__()
        # These are the lists that we combine (we don't combine everything,
        # because we can't combine white and black lists, and we don't
        # combine lists that return multiple results).
        # Note that DNSBL and URLBL are convenient labels, but DNSWL and
        # URLYL may also be also here.
        if os.path.exists(config_path):
            with open(config_path, "r") as fh:
                data = json.load(fh)
        else:
            data = {
              "COMBINED": "",
              "COMBINED_URL": "",
              "COMBINED_DNSBL": {},
              "COMBINED_DNSBL_REVERSE": {},
              "COMBINED_URLBL": {},
              "COMBINED_URLBL_REVERSE": {}
            }
        self.COMBINED = data["COMBINED"]
        self.COMBINED_URL = data["COMBINED_URL"]
        self.COMBINED_DNSBL = data["COMBINED_DNSBL"]
        self.COMBINED_DNSBL_REVERSE = data["COMBINED_DNSBL_REVERSE"]
        self.COMBINED_DNSBL_REVERSE_VALUES = \
            self.COMBINED_DNSBL_REVERSE.values()
        self.COMBINED_URLBL = data["COMBINED_URLBL"]
        self.COMBINED_URLBL_REVERSE = data["COMBINED_URLBL_REVERSE"]
        self.COMBINED_URLBL_REVERSE_VALUES = \
            self.COMBINED_URLBL_REVERSE.values()

    def lookup(self, question, qtype="A", ctype="IN", exact=False):
        """Do an actual lookup.  'question' should be the hostname or IP
        to query, and 'qType' should be the type of record to get
        (e.g. TXT, A, AAAA, PTR).

        If the lookup is within a domain that is handled by a combined
        list, then re-write the query so that it queries that list instead.
        We rely on the parent class caching the results so that when
        multiple lists in the combined system are queried, all but the
        first of these is pulled from the cache.

        When the question is not for one of the lists handled by
        a combined list the result is exactly the same as provided by
        Cache.lookup.
        When the question is one for one of the lists handled by
        a combined list, but the result indicates that the address is
        not listed, the result is an empty list.
        When the question is one for one of the lists handled by
        a combined list, and the result indicates that the address is
        listed, the result is always ["127.0.0.2"] - it is *not* the
        raw combined result.
        """
        logger = logging.getLogger("se-dns")
        rewrite_answer = None
        reverse_dict = None

        # XXX It would be better if this worked with any naming scheme.
        # Our lists always have 4 labels.
        # E.g: list1.dnsbl.example.com

        question_split = question.split(".")
        original_list = ".".join(question_split[-4:])
        address = ".".join(question_split[:-4])

        if (original_list in self.COMBINED_DNSBL_REVERSE_VALUES and
                self.COMBINED):
            logger.debug("Rewriting %s to use combined list.", question)
            rewrite_answer = original_list
            question = address + "." + self.COMBINED
            reverse_dict = self.COMBINED_DNSBL_REVERSE
        elif (original_list in self.COMBINED_URLBL_REVERSE_VALUES
              and self.COMBINED_URL):
            logger.debug("Rewriting %s to use url combined list.", question)
            rewrite_answer = original_list
            question = address + "." + self.COMBINED_URL
            reverse_dict = self.COMBINED_URLBL_REVERSE

        logger.debug("Looking up %s: %s", qtype, question)
        result = super(_DNSCache, self).lookup(question, qtype, ctype, exact)

        if rewrite_answer and result:
            for answer in result:
                if reverse_dict.get(answer) == rewrite_answer:
                    logger.debug("Converting %s from %s to ['127.0.0.2'] "
                                 "from %s.%s.", result, question, address,
                                 rewrite_answer)
                    result = ["127.0.0.2"]
                    break
            else:
                logger.debug("Ignoring %s from %s w.r.t. %s.%s", result,
                             question, address, rewrite_answer)
                result = []
        return result


# Central cache for all modules.
_DNS_CACHE = _DNSCache()


class DNSCache(object):
    """Proxy to the real DNSCache that allows per module timeouts."""

    def __init__(self, dnstimeout=10):
        self.dnsTimeout = dnstimeout
        self.COMBINED_DNSBL = _DNS_CACHE.COMBINED_DNSBL
        self.COMBINED_DNSBL_REVERSE = _DNS_CACHE.COMBINED_DNSBL_REVERSE
        self.COMBINED_DNSBL_REVERSE_VALUES = \
            _DNS_CACHE.COMBINED_DNSBL_REVERSE_VALUES
        self.COMBINED_URLBL = _DNS_CACHE.COMBINED_URLBL
        self.COMBINED_URLBL_REVERSE = _DNS_CACHE.COMBINED_URLBL_REVERSE
        self.COMBINED_URLBL_REVERSE_VALUES = \
            _DNS_CACHE.COMBINED_URLBL_REVERSE_VALUES

    def lookup(self, question, qtype="A", ctype="IN", exact=False):
        """Like Cache.lookup()"""
        # XXX This is not thread-safe
        _DNS_CACHE.queryObj.lifetime = self.dnsTimeout
        return _DNS_CACHE.lookup(question, qtype, ctype, exact)



#! /usr/bin/env python

from __future__ import absolute_import

import se_dns
import distutils.core

REQUIRES = ["dnspython"]
DESCRIPTION = """Simple dns tools."""

CLASSIFIERS = [
    "Operating System :: POSIX",
    "Programming Language :: Python",
    "Intended Audience :: System Administrators",
    "Development Status :: 5 - Production/Stable",
    "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
]

distutils.core.setup(
    name='se-dns',
    description=DESCRIPTION,
    author="SpamExperts",
    version=se_dns.__version__,
    license='GPL',
    platforms='POSIX',
    keywords='server',
    classifiers=CLASSIFIERS,
    # scripts=[],
    install_requires=REQUIRES,
    packages=[
        'se_dns',
    ],
)
